use bytes::{Buf, BufMut};

use crate::packet_encodable::PacketEncodable;

/// Struct holding an integer position in the world. Values are bound as follows:
/// -33554432 <= x <= 33554431
/// -2048 <= y <= 2047
/// -33554432 <= z <= 33554431
// The _phantom is used to stop construction of the struct outside of the new() function. Not for
// future api compatibility, but to avoid accidental instantiation of invalid positions.
#[allow(clippy::manual_non_exhaustive)]
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub struct Position {
    pub x: i32,
    pub y: i16,
    pub z: i32,
    _phantom: (),
}

impl Position {
    pub fn new(x: i32, y: i16, z: i32) -> Self {
        // TODO: Validate x, y, z are in range
        Self {
            x,
            y,
            z,
            _phantom: (),
        }
    }
}

impl PacketEncodable for Position {
    fn encode_packet(
        &self,
        buf: &mut impl BufMut,
    ) -> Result<(), crate::packet_encodable::PacketEncodeError> {
        let stuffed: i64 = ((self.x as i64 & 0x3FFFFFF) << 38)
            | ((self.z as i64 & 0x3FFFFFF) << 12)
            | (self.y as i64 & 0xFFF);
        stuffed.encode_packet(buf)
    }

    fn decode_packet(
        buf: &mut impl Buf,
    ) -> Result<Self, crate::packet_encodable::PacketDecodeError> {
        let stuffed = i64::decode_packet(buf)?;
        Ok(Position {
            x: (stuffed >> 38) as i32,
            y: (stuffed << 52 >> 52) as i16,
            z: (stuffed << 26 >> 38) as i32,
            _phantom: (),
        })
    }
}
