use bytes::{Buf, BufMut};
use potato_protocol_derive::{Packet, PacketEncodable};

use crate::packet_encodable::{PacketDecodeError, PacketEncodable, PacketEncodeError};

#[derive(Debug, Packet)]
#[packet(configuration_id = crate::ids::configuration::serverbound::CLIENT_INFORMATION)]
pub struct ClientInformationPacket {
    locale: String,
    view_distance: i8,
    chat_mode: ChatMode,
    chat_colors: bool,
    displayed_skin_parts: DisplayedSkinParts,
    main_hand: MainHand,
    enable_text_filtering: bool,
    allow_server_listings: bool,
    particle_status: ParticleStatus,
}

#[derive(Debug, PacketEncodable)]
pub enum ChatMode {
    #[packet(id = 0)]
    Enabled,
    #[packet(id = 1)]
    CommandsOnly,
    #[packet(id = 2)]
    Hidden,
}

#[derive(Debug, PacketEncodable)]
pub enum MainHand {
    #[packet(id = 0)]
    Left,
    #[packet(id = 1)]
    Right,
}

#[derive(Debug, PacketEncodable)]
pub enum ParticleStatus {
    #[packet(id = 0)]
    All,
    #[packet(id = 1)]
    Decreased,
    #[packet(id = 2)]
    Minimal,
}

#[derive(Debug)]
pub struct DisplayedSkinParts {
    pub cape: bool,
    pub jacket: bool,
    pub left_sleeve: bool,
    pub right_sleeve: bool,
    pub left_pants_leg: bool,
    pub right_pants_leg: bool,
    pub hat: bool,
}

impl PacketEncodable for DisplayedSkinParts {
    fn encode_packet(&self, buf: &mut impl BufMut) -> Result<(), PacketEncodeError> {
        let value = (self.cape as u8)
            | (self.jacket as u8) << 1
            | (self.left_sleeve as u8) << 2
            | (self.right_sleeve as u8) << 3
            | (self.left_pants_leg as u8) << 4
            | (self.right_pants_leg as u8) << 5
            | (self.hat as u8) << 6;

        value.encode_packet(buf)
    }

    fn decode_packet(buf: &mut impl Buf) -> Result<Self, PacketDecodeError> {
        let value = u8::decode_packet(buf)?;

        Ok(Self {
            cape: value & 0x01 != 0,
            jacket: value & 0x02 != 0,
            left_sleeve: value & 0x04 != 0,
            right_sleeve: value & 0x08 != 0,
            left_pants_leg: value & 0x10 != 0,
            right_pants_leg: value & 0x20 != 0,
            hat: value & 0x40 != 0,
        })
    }
}
